    make_pair(L'\U000000AC', L"\\lnot"),
    make_pair(L'\U000000B1', L"\\pm"),
    make_pair(L'\U000000C5', L"\\AA"),
    make_pair(L'\U000000D7', L"\\times"),
    make_pair(L'\U000000F7', L"\\div"),
    make_pair(L'\U00000393', L"\\Gamma"),
    make_pair(L'\U00000394', L"\\Delta"),
    make_pair(L'\U00000398', L"\\Theta"),
    make_pair(L'\U0000039B', L"\\Lambda"),
    make_pair(L'\U0000039E', L"\\Xi"),
    make_pair(L'\U000003A0', L"\\Pi"),
    make_pair(L'\U000003A3', L"\\Sigma"),
    make_pair(L'\U000003A5', L"\\Upsilon"),
    make_pair(L'\U000003A6', L"\\Phi"),
    make_pair(L'\U000003A8', L"\\Psi"),
    make_pair(L'\U000003A9', L"\\Omega"),
    make_pair(L'\U000003B1', L"\\alpha"),
    make_pair(L'\U000003B2', L"\\beta"),
    make_pair(L'\U000003B3', L"\\gamma"),
    make_pair(L'\U000003B4', L"\\delta"),
    make_pair(L'\U000003B5', L"\\varepsilon"),
    make_pair(L'\U000003B6', L"\\zeta"),
    make_pair(L'\U000003B7', L"\\eta"),
    make_pair(L'\U000003B8', L"\\theta"),
    make_pair(L'\U000003B9', L"\\iota"),
    make_pair(L'\U000003BA', L"\\kappa"),
    make_pair(L'\U000003BB', L"\\lambda"),
    make_pair(L'\U000003BC', L"\\mu"),
    make_pair(L'\U000003BD', L"\\nu"),
    make_pair(L'\U000003BE', L"\\xi"),
    make_pair(L'\U000003C0', L"\\pi"),
    make_pair(L'\U000003C1', L"\\rho"),
    make_pair(L'\U000003C2', L"\\varsigma"),
    make_pair(L'\U000003C3', L"\\sigma"),
    make_pair(L'\U000003C4', L"\\tau"),
    make_pair(L'\U000003C5', L"\\upsilon"),
    make_pair(L'\U000003C6', L"\\varphi"),
    make_pair(L'\U000003C7', L"\\chi"),
    make_pair(L'\U000003C8', L"\\psi"),
    make_pair(L'\U000003C9', L"\\omega"),
    make_pair(L'\U000003D1', L"\\vartheta"),
    make_pair(L'\U000003D5', L"\\phi"),
    make_pair(L'\U000003D6', L"\\varpi"),
    make_pair(L'\U000003DD', L"\\digamma"),
    make_pair(L'\U000003F0', L"\\varkappa"),
    make_pair(L'\U000003F1', L"\\varrho"),
    make_pair(L'\U000003F5', L"\\epsilon"),
    make_pair(L'\U000003F6', L"\\backepsilon"),
    make_pair(L'\U00002020', L"\\dagger"),
    make_pair(L'\U00002021', L"\\ddagger"),
    make_pair(L'\U00002022', L"\\bullet"),
    make_pair(L'\U00002026', L"\\dots"),
    make_pair(L'\U00002032', L"\\prime"),
    make_pair(L'\U00002035', L"\\backprime"),
    make_pair(L'\U00002190', L"\\leftarrow"),
    make_pair(L'\U00002191', L"\\uparrow"),
    make_pair(L'\U00002192', L"\\rightarrow"),
    make_pair(L'\U00002193', L"\\downarrow"),
    make_pair(L'\U00002194', L"\\leftrightarrow"),
    make_pair(L'\U00002195', L"\\updownarrow"),
    make_pair(L'\U00002196', L"\\nwarrow"),
    make_pair(L'\U00002197', L"\\nearrow"),
    make_pair(L'\U00002198', L"\\searrow"),
    make_pair(L'\U00002199', L"\\swarrow"),
    make_pair(L'\U0000219A', L"\\nleftarrow"),
    make_pair(L'\U0000219B', L"\\nrightarrow"),
    make_pair(L'\U0000219D', L"\\rightsquigarrow"),
    make_pair(L'\U0000219E', L"\\twoheadleftarrow"),
    make_pair(L'\U000021A0', L"\\twoheadrightarrow"),
    make_pair(L'\U000021A2', L"\\leftarrowtail"),
    make_pair(L'\U000021A3', L"\\rightarrowtail"),
    make_pair(L'\U000021A6', L"\\mapsto"),
    make_pair(L'\U000021A9', L"\\hookleftarrow"),
    make_pair(L'\U000021AA', L"\\hookrightarrow"),
    make_pair(L'\U000021AB', L"\\looparrowleft"),
    make_pair(L'\U000021AC', L"\\looparrowright"),
    make_pair(L'\U000021AD', L"\\leftrightsquigarrow"),
    make_pair(L'\U000021AE', L"\\nleftrightarrow"),
    make_pair(L'\U000021B0', L"\\Lsh"),
    make_pair(L'\U000021B1', L"\\Rsh"),
    make_pair(L'\U000021B6', L"\\curvearrowleft"),
    make_pair(L'\U000021B7', L"\\curvearrowright"),
    make_pair(L'\U000021BA', L"\\circlearrowleft"),
    make_pair(L'\U000021BB', L"\\circlearrowright"),
    make_pair(L'\U000021BC', L"\\leftharpoonup"),
    make_pair(L'\U000021BD', L"\\leftharpoondown"),
    make_pair(L'\U000021BE', L"\\upharpoonright"),
    make_pair(L'\U000021BF', L"\\upharpoonleft"),
    make_pair(L'\U000021C0', L"\\rightharpoonup"),
    make_pair(L'\U000021C1', L"\\rightharpoondown"),
    make_pair(L'\U000021C2', L"\\downharpoonright"),
    make_pair(L'\U000021C3', L"\\downharpoonleft"),
    make_pair(L'\U000021C4', L"\\rightleftarrows"),
    make_pair(L'\U000021C6', L"\\leftrightarrows"),
    make_pair(L'\U000021C7', L"\\leftleftarrows"),
    make_pair(L'\U000021C8', L"\\upuparrows"),
    make_pair(L'\U000021C9', L"\\rightrightarrows"),
    make_pair(L'\U000021CA', L"\\downdownarrows"),
    make_pair(L'\U000021CB', L"\\leftrightharpoons"),
    make_pair(L'\U000021CC', L"\\rightleftharpoons"),
    make_pair(L'\U000021CD', L"\\nLeftarrow"),
    make_pair(L'\U000021CE', L"\\nLeftrightarrow"),
    make_pair(L'\U000021CF', L"\\nRightarrow"),
    make_pair(L'\U000021D0', L"\\Leftarrow"),
    make_pair(L'\U000021D1', L"\\Uparrow"),
    make_pair(L'\U000021D2', L"\\Rightarrow"),
    make_pair(L'\U000021D3', L"\\Downarrow"),
    make_pair(L'\U000021D4', L"\\Leftrightarrow"),
    make_pair(L'\U000021D5', L"\\Updownarrow"),
    make_pair(L'\U000021DA', L"\\Lleftarrow"),
    make_pair(L'\U000021DB', L"\\Rrightarrow"),
    make_pair(L'\U000021DD', L"\\leadsto"),
    make_pair(L'\U00002200', L"\\forall"),
    make_pair(L'\U00002201', L"\\complement"),
    make_pair(L'\U00002203', L"\\exists"),
    make_pair(L'\U00002204', L"\\nexists"),
    make_pair(L'\U00002207', L"\\nabla"),
    make_pair(L'\U00002208', L"\\in"),
    make_pair(L'\U00002209', L"\\notin"),
    make_pair(L'\U0000220B', L"\\ni"),
    make_pair(L'\U0000220F', L"\\prod"),
    make_pair(L'\U00002210', L"\\coprod"),
    make_pair(L'\U00002211', L"\\sum"),
    make_pair(L'\U00002213', L"\\mp"),
    make_pair(L'\U00002214', L"\\dotplus"),
    make_pair(L'\U00002218', L"\\circ"),
    make_pair(L'\U0000221A', L"\\surd"),
    make_pair(L'\U0000221D', L"\\propto"),
    make_pair(L'\U00002220', L"\\angle"),
    make_pair(L'\U00002221', L"\\measuredangle"),
    make_pair(L'\U00002222', L"\\sphericalangle"),
    make_pair(L'\U00002224', L"\\nmid"),
    make_pair(L'\U00002225', L"\\parallel"),
    make_pair(L'\U00002226', L"\\nparallel"),
    make_pair(L'\U00002227', L"\\wedge"),
    make_pair(L'\U00002228', L"\\vee"),
    make_pair(L'\U00002229', L"\\cap"),
    make_pair(L'\U0000222A', L"\\cup"),
    make_pair(L'\U0000222B', L"\\int"),
    make_pair(L'\U0000222C', L"\\iint"),
    make_pair(L'\U0000222D', L"\\iiint"),
    make_pair(L'\U0000222E', L"\\oint"),
    make_pair(L'\U00002234', L"\\therefore"),
    make_pair(L'\U00002235', L"\\because"),
    make_pair(L'\U0000223C', L"\\sim"),
    make_pair(L'\U0000223D', L"\\backsim"),
    make_pair(L'\U00002240', L"\\wr"),
    make_pair(L'\U00002241', L"\\nsim"),
    make_pair(L'\U00002242', L"\\eqsim"),
    make_pair(L'\U00002243', L"\\simeq"),
    make_pair(L'\U00002245', L"\\cong"),
    make_pair(L'\U00002247', L"\\ncong"),
    make_pair(L'\U00002248', L"\\approx"),
    make_pair(L'\U0000224A', L"\\approxeq"),
    make_pair(L'\U0000224E', L"\\Bumpeq"),
    make_pair(L'\U0000224F', L"\\bumpeq"),
    make_pair(L'\U00002250', L"\\doteq"),
    make_pair(L'\U00002251', L"\\doteqdot"),
    make_pair(L'\U00002252', L"\\fallingdotseq"),
    make_pair(L'\U00002253', L"\\risingdotseq"),
    make_pair(L'\U00002256', L"\\eqcirc"),
    make_pair(L'\U00002257', L"\\circeq"),
    make_pair(L'\U0000225C', L"\\triangleq"),
    make_pair(L'\U00002260', L"\\neq"),
    make_pair(L'\U00002261', L"\\equiv"),
    make_pair(L'\U00002264', L"\\leq"),
    make_pair(L'\U00002265', L"\\geq"),
    make_pair(L'\U00002266', L"\\leqq"),
    make_pair(L'\U00002267', L"\\geqq"),
    make_pair(L'\U00002268', L"\\lneqq"),
    make_pair(L'\U00002269', L"\\gneqq"),
    make_pair(L'\U0000226A', L"\\ll"),
    make_pair(L'\U0000226B', L"\\gg"),
    make_pair(L'\U0000226C', L"\\between"),
    make_pair(L'\U0000226E', L"\\nless"),
    make_pair(L'\U0000226F', L"\\ngtr"),
    make_pair(L'\U00002270', L"\\nleq"),
    make_pair(L'\U00002271', L"\\ngeq"),
    make_pair(L'\U00002272', L"\\lesssim"),
    make_pair(L'\U00002273', L"\\gtrsim"),
    make_pair(L'\U00002276', L"\\lessgtr"),
    make_pair(L'\U00002277', L"\\gtrless"),
    make_pair(L'\U0000227A', L"\\prec"),
    make_pair(L'\U0000227B', L"\\succ"),
    make_pair(L'\U0000227C', L"\\preccurlyeq"),
    make_pair(L'\U0000227D', L"\\succcurlyeq"),
    make_pair(L'\U0000227E', L"\\precsim"),
    make_pair(L'\U0000227F', L"\\succsim"),
    make_pair(L'\U00002280', L"\\nprec"),
    make_pair(L'\U00002281', L"\\nsucc"),
    make_pair(L'\U00002282', L"\\subset"),
    make_pair(L'\U00002283', L"\\supset"),
    make_pair(L'\U00002286', L"\\subseteq"),
    make_pair(L'\U00002287', L"\\supseteq"),
    make_pair(L'\U00002288', L"\\nsubseteq"),
    make_pair(L'\U00002289', L"\\nsupseteq"),
    make_pair(L'\U0000228A', L"\\subsetneq"),
    make_pair(L'\U0000228B', L"\\supsetneq"),
    make_pair(L'\U0000228E', L"\\uplus"),
    make_pair(L'\U0000228F', L"\\sqsubset"),
    make_pair(L'\U00002290', L"\\sqsupset"),
    make_pair(L'\U00002291', L"\\sqsubseteq"),
    make_pair(L'\U00002292', L"\\sqsupseteq"),
    make_pair(L'\U00002293', L"\\sqcap"),
    make_pair(L'\U00002294', L"\\sqcup"),
    make_pair(L'\U00002295', L"\\oplus"),
    make_pair(L'\U00002296', L"\\ominus"),
    make_pair(L'\U00002297', L"\\otimes"),
    make_pair(L'\U00002298', L"\\oslash"),
    make_pair(L'\U00002299', L"\\odot"),
    make_pair(L'\U0000229A', L"\\circledcirc"),
    make_pair(L'\U0000229B', L"\\circledast"),
    make_pair(L'\U0000229D', L"\\circleddash"),
    make_pair(L'\U0000229E', L"\\boxplus"),
    make_pair(L'\U0000229F', L"\\boxminus"),
    make_pair(L'\U000022A0', L"\\boxtimes"),
    make_pair(L'\U000022A1', L"\\boxdot"),
    make_pair(L'\U000022A2', L"\\vdash"),
    make_pair(L'\U000022A3', L"\\dashv"),
    make_pair(L'\U000022A4', L"\\top"),
    make_pair(L'\U000022A5', L"\\bot"),
    make_pair(L'\U000022A7', L"\\models"),
    make_pair(L'\U000022A8', L"\\vDash"),
    make_pair(L'\U000022A9', L"\\Vdash"),
    make_pair(L'\U000022AA', L"\\Vvdash"),
    make_pair(L'\U000022AC', L"\\nvdash"),
    make_pair(L'\U000022AD', L"\\nvDash"),
    make_pair(L'\U000022AE', L"\\nVdash"),
    make_pair(L'\U000022AF', L"\\nVDash"),
    make_pair(L'\U000022B2', L"\\lhd"),
    make_pair(L'\U000022B3', L"\\rhd"),
    make_pair(L'\U000022B4', L"\\unlhd"),
    make_pair(L'\U000022B5', L"\\unrhd"),
    make_pair(L'\U000022B8', L"\\multimap"),
    make_pair(L'\U000022BA', L"\\intercal"),
    make_pair(L'\U000022BB', L"\\veebar"),
    make_pair(L'\U000022C0', L"\\bigwedge"),
    make_pair(L'\U000022C1', L"\\bigvee"),
    make_pair(L'\U000022C2', L"\\bigcap"),
    make_pair(L'\U000022C3', L"\\bigcup"),
    make_pair(L'\U000022C4', L"\\diamond"),
    make_pair(L'\U000022C5', L"\\cdot"),
    make_pair(L'\U000022C6', L"\\star"),
    make_pair(L'\U000022C7', L"\\divideontimes"),
    make_pair(L'\U000022C8', L"\\bowtie"),
    make_pair(L'\U000022C9', L"\\ltimes"),
    make_pair(L'\U000022CA', L"\\rtimes"),
    make_pair(L'\U000022CB', L"\\leftthreetimes"),
    make_pair(L'\U000022CC', L"\\rightthreetimes"),
    make_pair(L'\U000022CD', L"\\backsimeq"),
    make_pair(L'\U000022CE', L"\\curlyvee"),
    make_pair(L'\U000022CF', L"\\curlywedge"),
    make_pair(L'\U000022D0', L"\\Subset"),
    make_pair(L'\U000022D1', L"\\Supset"),
    make_pair(L'\U000022D2', L"\\Cap"),
    make_pair(L'\U000022D3', L"\\Cup"),
    make_pair(L'\U000022D4', L"\\pitchfork"),
    make_pair(L'\U000022D6', L"\\lessdot"),
    make_pair(L'\U000022D7', L"\\gtrdot"),
    make_pair(L'\U000022D8', L"\\lll"),
    make_pair(L'\U000022D9', L"\\ggg"),
    make_pair(L'\U000022DA', L"\\lesseqgtr"),
    make_pair(L'\U000022DB', L"\\gtreqless"),
    make_pair(L'\U000022DE', L"\\curlyeqprec"),
    make_pair(L'\U000022DF', L"\\curlyeqsucc"),
    make_pair(L'\U000022E6', L"\\lnsim"),
    make_pair(L'\U000022E7', L"\\gnsim"),
    make_pair(L'\U000022E8', L"\\precnsim"),
    make_pair(L'\U000022E9', L"\\succnsim"),
    make_pair(L'\U000022EA', L"\\ntriangleleft"),
    make_pair(L'\U000022EB', L"\\ntriangleright"),
    make_pair(L'\U000022EC', L"\\ntrianglelefteq"),
    make_pair(L'\U000022ED', L"\\ntrianglerighteq"),
    make_pair(L'\U000022EE', L"\\vdots"),
    make_pair(L'\U000022EF', L"\\cdots"),
    make_pair(L'\U000022F1', L"\\ddots"),
    make_pair(L'\U00002305', L"\\barwedge"),
    make_pair(L'\U00002306', L"\\doublebarwedge"),
    make_pair(L'\U00002308', L"\\lceil"),
    make_pair(L'\U00002309', L"\\rceil"),
    make_pair(L'\U0000230A', L"\\lfloor"),
    make_pair(L'\U0000230B', L"\\rfloor"),
    make_pair(L'\U0000231C', L"\\ulcorner"),
    make_pair(L'\U0000231D', L"\\urcorner"),
    make_pair(L'\U0000231E', L"\\llcorner"),
    make_pair(L'\U0000231F', L"\\lrcorner"),
    make_pair(L'\U00002322', L"\\frown"),
    make_pair(L'\U00002323', L"\\smile"),
    make_pair(L'\U00002329', L"\\langle"),
    make_pair(L'\U0000232A', L"\\rangle"),
    make_pair(L'\U000025A1', L"\\square"),
    make_pair(L'\U000025B3', L"\\triangle"),
    make_pair(L'\U000025B4', L"\\blacktriangle"),
    make_pair(L'\U000025B5', L"\\vartriangle"),
    make_pair(L'\U000025B6', L"\\blacktriangleright"),
    make_pair(L'\U000025B9', L"\\triangleright"),
    make_pair(L'\U000025BD', L"\\bigtriangledown"),
    make_pair(L'\U000025BE', L"\\blacktriangledown"),
    make_pair(L'\U000025BF', L"\\triangledown"),
    make_pair(L'\U000025C0', L"\\blacktriangleleft"),
    make_pair(L'\U000025C3', L"\\triangleleft"),
    make_pair(L'\U000025CA', L"\\lozenge"),
    make_pair(L'\U000025EF', L"\\bigcirc"),
    make_pair(L'\U000025FC', L"\\blacksquare"),
    make_pair(L'\U00002605', L"\\bigstar"),
    make_pair(L'\U00002660', L"\\spadesuit"),
    make_pair(L'\U00002663', L"\\clubsuit"),
    make_pair(L'\U00002665', L"\\heartsuit"),
    make_pair(L'\U00002666', L"\\diamondsuit"),
    make_pair(L'\U0000266D', L"\\flat"),
    make_pair(L'\U0000266E', L"\\natural"),
    make_pair(L'\U0000266F', L"\\sharp"),
    make_pair(L'\U00002713', L"\\checkmark"),
    make_pair(L'\U0000290E', L"\\dashleftarrow"),
    make_pair(L'\U0000290F', L"\\dashrightarrow"),
    make_pair(L'\U000029EB', L"\\blacklozenge"),
    make_pair(L'\U00002A00', L"\\bigodot"),
    make_pair(L'\U00002A01', L"\\bigoplus"),
    make_pair(L'\U00002A02', L"\\bigotimes"),
    make_pair(L'\U00002A04', L"\\biguplus"),
    make_pair(L'\U00002A06', L"\\bigsqcup"),
    make_pair(L'\U00002A0C', L"\\iiiint"),
    make_pair(L'\U00002A3F', L"\\amalg"),
    make_pair(L'\U00002A7D', L"\\leqslant"),
    make_pair(L'\U00002A7E', L"\\geqslant"),
    make_pair(L'\U00002A85', L"\\lessapprox"),
    make_pair(L'\U00002A86', L"\\gtrapprox"),
    make_pair(L'\U00002A89', L"\\lnapprox"),
    make_pair(L'\U00002A8A', L"\\gnapprox"),
    make_pair(L'\U00002A8B', L"\\lesseqqgtr"),
    make_pair(L'\U00002A8C', L"\\gtreqqless"),
    make_pair(L'\U00002A95', L"\\eqslantless"),
    make_pair(L'\U00002A96', L"\\eqslantgtr"),
    make_pair(L'\U00002AAF', L"\\preceq"),
    make_pair(L'\U00002AB0', L"\\succeq"),
    make_pair(L'\U00002AB5', L"\\precneqq"),
    make_pair(L'\U00002AB6', L"\\succneqq"),
    make_pair(L'\U00002AB7', L"\\precapprox"),
    make_pair(L'\U00002AB8', L"\\succapprox"),
    make_pair(L'\U00002AB9', L"\\precnapprox"),
    make_pair(L'\U00002ABA', L"\\succnapprox"),
    make_pair(L'\U00002AC5', L"\\subseteqq"),
    make_pair(L'\U00002AC6', L"\\supseteqq"),
    make_pair(L'\U00002ACB', L"\\subsetneqq"),
    make_pair(L'\U00002ACC', L"\\supsetneqq")
